/**
 * @copyright   2018-2019, Miles Johnson
 * @license     https://opensource.org/licenses/MIT
 */
import { types as t } from '@babel/core';
import { convertSymbolFromSource } from './convertTSToPropTypes';
import getTypeName from './getTypeName';
import { createCall, createMember, hasCustomPropTypeSuffix, isReactTypeMatch, wrapIsRequired } from './propTypes';
const NATIVE_BUILT_INS = ['Date', 'Error', 'RegExp', 'Map', 'WeakMap', 'Set', 'WeakSet', 'Promise'];

function convert(type, state, depth) {
  const {
    reactImportedName,
    propTypes
  } = state;
  const propTypesImportedName = propTypes.defaultImport;
  const isMaxDepth = depth >= state.options.maxDepth;

  if (t.isTSParenthesizedType(type)) {
    type = type.typeAnnotation;
  }

  state.propTypes.count += 1;

  if (t.isTSAnyKeyword(type)) {
    return createMember(t.identifier('any'), propTypesImportedName);
  } else if (t.isTSStringKeyword(type)) {
    return createMember(t.identifier('string'), propTypesImportedName);
  } else if (t.isTSNumberKeyword(type)) {
    return createMember(t.identifier('number'), propTypesImportedName);
  } else if (t.isTSBooleanKeyword(type)) {
    return createMember(t.identifier('bool'), propTypesImportedName);
  } else if (t.isTSSymbolKeyword(type)) {
    return createMember(t.identifier('symbol'), propTypesImportedName);
  } else if (t.isTSObjectKeyword(type)) {
    return createMember(t.identifier('object'), propTypesImportedName);
  } else if (t.isTSNullKeyword(type)) {
    return createCall(t.identifier('oneOf'), [t.arrayExpression([t.nullLiteral()])], propTypesImportedName);
  } else if (t.isTSLiteralType(type)) {
    return createCall(t.identifier('oneOf'), [t.arrayExpression([type.literal])], propTypesImportedName);
  } else if (t.isTSFunctionType(type)) {
    return createMember(t.identifier('func'), propTypesImportedName);
  } else if (t.isTSTypeReference(type)) {
    const name = getTypeName(type.typeName);

    if (isReactTypeMatch(name, 'ReactText', reactImportedName) || isReactTypeMatch(name, 'ReactNode', reactImportedName) || isReactTypeMatch(name, 'ReactType', reactImportedName)) {
      return createMember(t.identifier('node'), propTypesImportedName);
    } else if (isReactTypeMatch(name, 'ComponentType', reactImportedName) || isReactTypeMatch(name, 'ComponentClass', reactImportedName) || isReactTypeMatch(name, 'StatelessComponent', reactImportedName)) {
      return createMember(t.identifier('func'), propTypesImportedName);
    } else if (isReactTypeMatch(name, 'Element', 'JSX') || isReactTypeMatch(name, 'ReactElement', reactImportedName) || isReactTypeMatch(name, 'ComponentElement', reactImportedName) || isReactTypeMatch(name, 'FunctionComponentElement', reactImportedName) || isReactTypeMatch(name, 'DOMElement', reactImportedName) || isReactTypeMatch(name, 'SFCElement', reactImportedName)) {
      return createMember(t.identifier('element'), propTypesImportedName);
    } else if (isReactTypeMatch(name, 'Ref', reactImportedName)) {
      return createCall(t.identifier('oneOfType'), [t.arrayExpression([createMember(t.identifier('string'), propTypesImportedName), createMember(t.identifier('func'), propTypesImportedName), createMember(t.identifier('object'), propTypesImportedName)])], propTypesImportedName);
    } else if (name.endsWith('Handler')) {
      return createMember(t.identifier('func'), propTypesImportedName);
    } else if (name.endsWith('Event')) {
      return createMember(t.identifier('object'), propTypesImportedName);
    } else if (NATIVE_BUILT_INS.includes(name)) {
      return createCall(t.identifier('instanceOf'), [t.identifier(name)], propTypesImportedName);
    } else if (state.referenceTypes[name]) {
      return convert(state.referenceTypes[name], state, depth);
    } else if (hasCustomPropTypeSuffix(name, state.options.customPropTypeSuffixes)) {
      return t.identifier(name);
    } else if (state.typeChecker) {
      return convertSymbolFromSource(state.filePath, name, state);
    }

    return createMember(t.identifier('any'), propTypesImportedName);
  } else if (t.isTSArrayType(type)) {
    const args = convertArray([type.elementType], state, depth);
    return args.length > 0 ? createCall(t.identifier('arrayOf'), args, propTypesImportedName) : createMember(t.identifier('array'), propTypesImportedName);
  } else if (t.isTSTypeLiteral(type)) {
    if (type.members.length === 0 || isMaxDepth) {
      return createMember(t.identifier('object'), propTypesImportedName);
    } else if (type.members.length === 1 && t.isTSIndexSignature(type.members[0])) {
      const index = type.members[0];

      if (index.typeAnnotation && index.typeAnnotation.typeAnnotation) {
        const result = convert(index.typeAnnotation.typeAnnotation, state, depth);

        if (result) {
          return createCall(t.identifier('objectOf'), [result], propTypesImportedName);
        }
      }
    } else {
      return createCall(t.identifier('shape'), [t.objectExpression(convertListToProps(type.members.filter(member => t.isTSPropertySignature(member)), state, [], depth + 1))], propTypesImportedName);
    }
  } else if (t.isTSUnionType(type) || t.isTSIntersectionType(type)) {
    const isAllLiterals = type.types.every(param => t.isTSLiteralType(param));
    const containsAny = type.types.some(param => t.isTSAnyKeyword(param));
    let label;
    let args;

    if (isAllLiterals) {
      args = type.types.map(param => param.literal);
      label = t.identifier('oneOf');

      if (state.options.maxSize) {
        args = args.slice(0, state.options.maxSize);
      }
    } else if (containsAny) {
      return createMember(t.identifier('any'), propTypesImportedName);
    } else {
      args = convertArray(type.types, state, depth);
      label = t.identifier('oneOfType');
    }

    if (label && args.length > 0) {
      return createCall(label, [t.arrayExpression(args)], propTypesImportedName);
    }
  } else if (t.isTSInterfaceDeclaration(type)) {
    if (type.body.body.length === 0 || isMaxDepth) {
      return createMember(t.identifier('object'), propTypesImportedName);
    }

    return createCall(t.identifier('shape'), [t.objectExpression(convertListToProps(type.body.body.filter(property => t.isTSPropertySignature(property)), state, [], depth + 1))], propTypesImportedName);
  } else if (t.isTSTypeAliasDeclaration(type)) {
    return convert(type.typeAnnotation, state, depth);
  } else if (t.isTSIndexedAccessType(type)) {
    const {
      objectType,
      indexType
    } = type;

    if (t.isTSTypeReference(objectType) && t.isTSLiteralType(indexType)) {
      const ref = state.referenceTypes[objectType.typeName.name];
      let properties;

      if (t.isTSInterfaceDeclaration(ref)) {
        properties = ref.body.body;
      } else if (t.isTSTypeAliasDeclaration(ref) && t.isTSTypeLiteral(ref.typeAnnotation)) {
        properties = ref.typeAnnotation.members;
      } else {
        return null;
      }

      const property = properties.find(prop => t.isTSPropertySignature(prop) && prop.key.name === indexType.literal.value);
      return property ? convert(property.typeAnnotation.typeAnnotation, state, depth) : null;
    }
  } else if (t.isTSTypeQuery(type)) {
    return createMember(t.identifier('any'), propTypesImportedName);
  } else if (t.isTSTypeOperator(type) && type.operator === 'keyof') {
    return createMember(t.identifier('any'), propTypesImportedName);
  }

  state.propTypes.count -= 1;
  return null;
}

function mustBeOptional(type) {
  if (t.isTSUnionType(type)) {
    return type.types.some(value => t.isTSAnyKeyword(value) || t.isTSNullKeyword(value) || t.isTSUndefinedKeyword(value));
  }

  return false;
}

function convertArray(types, state, depth) {
  const propTypes = [];
  types.forEach(type => {
    const prop = convert(type, state, depth);

    if (prop) {
      propTypes.push(prop);
    }
  });
  return propTypes;
}

function convertListToProps(properties, state, defaultProps, depth) {
  const propTypes = [];
  let hasChildren = false;
  let size = 0;
  properties.some(property => {
    if (state.options.maxSize && size === state.options.maxSize) {
      return true;
    }

    if (!property.typeAnnotation) {
      return false;
    }

    const type = property.typeAnnotation.typeAnnotation;
    const propType = convert(type, state, depth);
    const {
      name
    } = property.key;

    if (propType) {
      propTypes.push(t.objectProperty(property.key, wrapIsRequired(propType, property.optional || defaultProps.includes(name) || mustBeOptional(type))));

      if (name === 'children') {
        hasChildren = true;
      }

      size += 1;
    }

    return false;
  });

  if (!hasChildren && depth === 0 && propTypes.length > 0 && state.options.implicitChildren) {
    propTypes.push(t.objectProperty(t.identifier('children'), createMember(t.identifier('node'), state.propTypes.defaultImport)));
  }

  return propTypes;
}

export default function convertToPropTypes(types, typeNames, state, defaultProps) {
  const properties = [];
  typeNames.forEach(typeName => {
    if (types[typeName]) {
      properties.push(...convertListToProps(types[typeName], state, defaultProps, 0));
    }
  });
  return properties;
}