/**
 * @copyright   2018-2019, Miles Johnson
 * @license     https://opensource.org/licenses/MIT
 */
import { types as t } from '@babel/core';
import ts from 'typescript';
import { createCall, createMember, hasCustomPropTypeSuffix, isReactTypeMatch } from './propTypes';
export function convert(type, state, depth) {
  const {
    reactImportedName,
    propTypes
  } = state;
  const propTypesImportedName = propTypes.defaultImport;
  const isMaxDepth = depth >= (state.options.maxDepth || 3);
  state.propTypes.count += 1;

  if (type.flags & ts.TypeFlags.Any) {
    return createMember(t.identifier('any'), propTypesImportedName);
  } else if (type.flags & ts.TypeFlags.StringLike) {
    return createMember(t.identifier('string'), propTypesImportedName);
  } else if (type.flags & ts.TypeFlags.NumberLike) {
    return createMember(t.identifier('number'), propTypesImportedName);
  } else if (type.flags & ts.TypeFlags.BooleanLike) {
    return createMember(t.identifier('bool'), propTypesImportedName);
  } else if (type.flags & ts.TypeFlags.ESSymbolLike) {
    return createMember(t.identifier('symbol'), propTypesImportedName);
  } else if (type.flags & ts.TypeFlags.Object) {
    const objType = type;

    if (objType.objectFlags & ts.ObjectFlags.Interface) {} else if (objType.objectFlags & ts.ObjectFlags.Reference) {} else if (objType.objectFlags & ts.ObjectFlags.Tuple) {} else if (objType.objectFlags & ts.ObjectFlags.Mapped) {} else if (objType.objectFlags & ts.ObjectFlags.Instantiated) {} else if (objType.objectFlags & ts.ObjectFlags.ObjectLiteral) {}

    return createMember(t.identifier('object'), propTypesImportedName);
  } else if (type.flags & ts.TypeFlags.Any) {
    return createMember(t.identifier('func'), propTypesImportedName);
  } else if (t.isTSTypeReference(type)) {
    const name = '';

    if (isReactTypeMatch(name, 'ReactText', reactImportedName) || isReactTypeMatch(name, 'ReactNode', reactImportedName) || isReactTypeMatch(name, 'ReactType', reactImportedName)) {
      return createMember(t.identifier('node'), propTypesImportedName);
    } else if (isReactTypeMatch(name, 'ComponentType', reactImportedName) || isReactTypeMatch(name, 'ComponentClass', reactImportedName) || isReactTypeMatch(name, 'StatelessComponent', reactImportedName)) {
      return createMember(t.identifier('func'), propTypesImportedName);
    } else if (isReactTypeMatch(name, 'Element', 'JSX') || isReactTypeMatch(name, 'ReactElement', reactImportedName) || isReactTypeMatch(name, 'ComponentElement', reactImportedName) || isReactTypeMatch(name, 'FunctionComponentElement', reactImportedName) || isReactTypeMatch(name, 'DOMElement', reactImportedName) || isReactTypeMatch(name, 'SFCElement', reactImportedName)) {
      return createMember(t.identifier('element'), propTypesImportedName);
    } else if (isReactTypeMatch(name, 'Ref', reactImportedName)) {
      return createCall(t.identifier('oneOfType'), [t.arrayExpression([createMember(t.identifier('string'), propTypesImportedName), createMember(t.identifier('func'), propTypesImportedName), createMember(t.identifier('object'), propTypesImportedName)])], propTypesImportedName);
    } else if (name.endsWith('Handler')) {
      return createMember(t.identifier('func'), propTypesImportedName);
    } else if (name.endsWith('Event')) {
      return createMember(t.identifier('object'), propTypesImportedName);
    } else if (hasCustomPropTypeSuffix(name, state.options.customPropTypeSuffixes)) {
      return t.identifier(name);
    } else if (state.typeChecker) {
      return convertSymbolFromSource(state.filePath, name, state);
    }

    return createMember(t.identifier('any'), propTypesImportedName);
  } else if (t.isTSArrayType(type)) {} else if (t.isTSTypeLiteral(type)) {
    if (type.members.length === 0 || isMaxDepth) {
      return createMember(t.identifier('object'), propTypesImportedName);
    } else if (type.members.length === 1 && t.isTSIndexSignature(type.members[0])) {} else {}
  } else if (t.isTSUnionType(type) || t.isTSIntersectionType(type)) {} else if (t.isTSInterfaceDeclaration(type)) {} else if (t.isTSTypeAliasDeclaration(type)) {}

  state.propTypes.count -= 1;
  return null;
}
export function convertSymbolFromSource(filePath, symbolName, state) {
  const program = state.typeProgram;
  const checker = state.typeChecker;
  const source = program.getSourceFile(filePath);

  if (!source) {
    return null;
  }

  let refNode = null;
  source.statements.some(node => {
    if (ts.isImportDeclaration(node) && node.importClause) {
      const {
        name,
        namedBindings
      } = node.importClause;

      if (name && name.text === symbolName) {
        refNode = name;
        return true;
      }

      if (!refNode && namedBindings && ts.isNamedImports(namedBindings)) {
        return namedBindings.elements.some(element => {
          if (element.name.text === symbolName) {
            refNode = element.name;
            return true;
          }

          return false;
        });
      }
    }

    return false;
  });

  if (!refNode) {
    return null;
  }

  const symbol = checker.getSymbolAtLocation(refNode);

  if (!symbol) {
    return null;
  }

  return null;
}