"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = convertToPropTypes;

var _core = require("@babel/core");

var _convertTSToPropTypes = require("./convertTSToPropTypes");

var _getTypeName = _interopRequireDefault(require("./getTypeName"));

var _propTypes = require("./propTypes");

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @copyright   2018-2019, Miles Johnson
 * @license     https://opensource.org/licenses/MIT
 */
const NATIVE_BUILT_INS = ['Date', 'Error', 'RegExp', 'Map', 'WeakMap', 'Set', 'WeakSet', 'Promise'];

function convert(type, state, depth) {
  const {
    reactImportedName,
    propTypes
  } = state;
  const propTypesImportedName = propTypes.defaultImport;
  const isMaxDepth = depth >= state.options.maxDepth;

  if (_core.types.isTSParenthesizedType(type)) {
    type = type.typeAnnotation;
  }

  state.propTypes.count += 1;

  if (_core.types.isTSAnyKeyword(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('any'), propTypesImportedName);
  } else if (_core.types.isTSStringKeyword(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('string'), propTypesImportedName);
  } else if (_core.types.isTSNumberKeyword(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('number'), propTypesImportedName);
  } else if (_core.types.isTSBooleanKeyword(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('bool'), propTypesImportedName);
  } else if (_core.types.isTSSymbolKeyword(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('symbol'), propTypesImportedName);
  } else if (_core.types.isTSObjectKeyword(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('object'), propTypesImportedName);
  } else if (_core.types.isTSNullKeyword(type)) {
    return (0, _propTypes.createCall)(_core.types.identifier('oneOf'), [_core.types.arrayExpression([_core.types.nullLiteral()])], propTypesImportedName);
  } else if (_core.types.isTSLiteralType(type)) {
    return (0, _propTypes.createCall)(_core.types.identifier('oneOf'), [_core.types.arrayExpression([type.literal])], propTypesImportedName);
  } else if (_core.types.isTSFunctionType(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('func'), propTypesImportedName);
  } else if (_core.types.isTSTypeReference(type)) {
    const name = (0, _getTypeName.default)(type.typeName);

    if ((0, _propTypes.isReactTypeMatch)(name, 'ReactText', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'ReactNode', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'ReactType', reactImportedName)) {
      return (0, _propTypes.createMember)(_core.types.identifier('node'), propTypesImportedName);
    } else if ((0, _propTypes.isReactTypeMatch)(name, 'ComponentType', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'ComponentClass', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'StatelessComponent', reactImportedName)) {
      return (0, _propTypes.createMember)(_core.types.identifier('func'), propTypesImportedName);
    } else if ((0, _propTypes.isReactTypeMatch)(name, 'Element', 'JSX') || (0, _propTypes.isReactTypeMatch)(name, 'ReactElement', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'ComponentElement', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'FunctionComponentElement', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'DOMElement', reactImportedName) || (0, _propTypes.isReactTypeMatch)(name, 'SFCElement', reactImportedName)) {
      return (0, _propTypes.createMember)(_core.types.identifier('element'), propTypesImportedName);
    } else if ((0, _propTypes.isReactTypeMatch)(name, 'Ref', reactImportedName)) {
      return (0, _propTypes.createCall)(_core.types.identifier('oneOfType'), [_core.types.arrayExpression([(0, _propTypes.createMember)(_core.types.identifier('string'), propTypesImportedName), (0, _propTypes.createMember)(_core.types.identifier('func'), propTypesImportedName), (0, _propTypes.createMember)(_core.types.identifier('object'), propTypesImportedName)])], propTypesImportedName);
    } else if (name.endsWith('Handler')) {
      return (0, _propTypes.createMember)(_core.types.identifier('func'), propTypesImportedName);
    } else if (name.endsWith('Event')) {
      return (0, _propTypes.createMember)(_core.types.identifier('object'), propTypesImportedName);
    } else if (NATIVE_BUILT_INS.includes(name)) {
      return (0, _propTypes.createCall)(_core.types.identifier('instanceOf'), [_core.types.identifier(name)], propTypesImportedName);
    } else if (state.referenceTypes[name]) {
      return convert(state.referenceTypes[name], state, depth);
    } else if ((0, _propTypes.hasCustomPropTypeSuffix)(name, state.options.customPropTypeSuffixes)) {
      return _core.types.identifier(name);
    } else if (state.typeChecker) {
      return (0, _convertTSToPropTypes.convertSymbolFromSource)(state.filePath, name, state);
    }

    return (0, _propTypes.createMember)(_core.types.identifier('any'), propTypesImportedName);
  } else if (_core.types.isTSArrayType(type)) {
    const args = convertArray([type.elementType], state, depth);
    return args.length > 0 ? (0, _propTypes.createCall)(_core.types.identifier('arrayOf'), args, propTypesImportedName) : (0, _propTypes.createMember)(_core.types.identifier('array'), propTypesImportedName);
  } else if (_core.types.isTSTypeLiteral(type)) {
    if (type.members.length === 0 || isMaxDepth) {
      return (0, _propTypes.createMember)(_core.types.identifier('object'), propTypesImportedName);
    } else if (type.members.length === 1 && _core.types.isTSIndexSignature(type.members[0])) {
      const index = type.members[0];

      if (index.typeAnnotation && index.typeAnnotation.typeAnnotation) {
        const result = convert(index.typeAnnotation.typeAnnotation, state, depth);

        if (result) {
          return (0, _propTypes.createCall)(_core.types.identifier('objectOf'), [result], propTypesImportedName);
        }
      }
    } else {
      return (0, _propTypes.createCall)(_core.types.identifier('shape'), [_core.types.objectExpression(convertListToProps(type.members.filter(member => _core.types.isTSPropertySignature(member)), state, [], depth + 1))], propTypesImportedName);
    }
  } else if (_core.types.isTSUnionType(type) || _core.types.isTSIntersectionType(type)) {
    const isAllLiterals = type.types.every(param => _core.types.isTSLiteralType(param));
    const containsAny = type.types.some(param => _core.types.isTSAnyKeyword(param));
    let label;
    let args;

    if (isAllLiterals) {
      args = type.types.map(param => param.literal);
      label = _core.types.identifier('oneOf');

      if (state.options.maxSize) {
        args = args.slice(0, state.options.maxSize);
      }
    } else if (containsAny) {
      return (0, _propTypes.createMember)(_core.types.identifier('any'), propTypesImportedName);
    } else {
      args = convertArray(type.types, state, depth);
      label = _core.types.identifier('oneOfType');
    }

    if (label && args.length > 0) {
      return (0, _propTypes.createCall)(label, [_core.types.arrayExpression(args)], propTypesImportedName);
    }
  } else if (_core.types.isTSInterfaceDeclaration(type)) {
    if (type.body.body.length === 0 || isMaxDepth) {
      return (0, _propTypes.createMember)(_core.types.identifier('object'), propTypesImportedName);
    }

    return (0, _propTypes.createCall)(_core.types.identifier('shape'), [_core.types.objectExpression(convertListToProps(type.body.body.filter(property => _core.types.isTSPropertySignature(property)), state, [], depth + 1))], propTypesImportedName);
  } else if (_core.types.isTSTypeAliasDeclaration(type)) {
    return convert(type.typeAnnotation, state, depth);
  } else if (_core.types.isTSIndexedAccessType(type)) {
    const {
      objectType,
      indexType
    } = type;

    if (_core.types.isTSTypeReference(objectType) && _core.types.isTSLiteralType(indexType)) {
      const ref = state.referenceTypes[objectType.typeName.name];
      let properties;

      if (_core.types.isTSInterfaceDeclaration(ref)) {
        properties = ref.body.body;
      } else if (_core.types.isTSTypeAliasDeclaration(ref) && _core.types.isTSTypeLiteral(ref.typeAnnotation)) {
        properties = ref.typeAnnotation.members;
      } else {
        return null;
      }

      const property = properties.find(prop => _core.types.isTSPropertySignature(prop) && prop.key.name === indexType.literal.value);
      return property ? convert(property.typeAnnotation.typeAnnotation, state, depth) : null;
    }
  } else if (_core.types.isTSTypeQuery(type)) {
    return (0, _propTypes.createMember)(_core.types.identifier('any'), propTypesImportedName);
  } else if (_core.types.isTSTypeOperator(type) && type.operator === 'keyof') {
    return (0, _propTypes.createMember)(_core.types.identifier('any'), propTypesImportedName);
  }

  state.propTypes.count -= 1;
  return null;
}

function mustBeOptional(type) {
  if (_core.types.isTSUnionType(type)) {
    return type.types.some(value => _core.types.isTSAnyKeyword(value) || _core.types.isTSNullKeyword(value) || _core.types.isTSUndefinedKeyword(value));
  }

  return false;
}

function convertArray(types, state, depth) {
  const propTypes = [];
  types.forEach(type => {
    const prop = convert(type, state, depth);

    if (prop) {
      propTypes.push(prop);
    }
  });
  return propTypes;
}

function convertListToProps(properties, state, defaultProps, depth) {
  const propTypes = [];
  let hasChildren = false;
  let size = 0;
  properties.some(property => {
    if (state.options.maxSize && size === state.options.maxSize) {
      return true;
    }

    if (!property.typeAnnotation) {
      return false;
    }

    const type = property.typeAnnotation.typeAnnotation;
    const propType = convert(type, state, depth);
    const {
      name
    } = property.key;

    if (propType) {
      propTypes.push(_core.types.objectProperty(property.key, (0, _propTypes.wrapIsRequired)(propType, property.optional || defaultProps.includes(name) || mustBeOptional(type))));

      if (name === 'children') {
        hasChildren = true;
      }

      size += 1;
    }

    return false;
  });

  if (!hasChildren && depth === 0 && propTypes.length > 0 && state.options.implicitChildren) {
    propTypes.push(_core.types.objectProperty(_core.types.identifier('children'), (0, _propTypes.createMember)(_core.types.identifier('node'), state.propTypes.defaultImport)));
  }

  return propTypes;
}

function convertToPropTypes(types, typeNames, state, defaultProps) {
  const properties = [];
  typeNames.forEach(typeName => {
    if (types[typeName]) {
      properties.push(...convertListToProps(types[typeName], state, defaultProps, 0));
    }
  });
  return properties;
}