"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;

var _helperPluginUtils = require("@babel/helper-plugin-utils");

var _helperModuleImports = require("@babel/helper-module-imports");

var _pluginSyntaxTypescript = _interopRequireDefault(require("@babel/plugin-syntax-typescript"));

var _core = require("@babel/core");

var _addToClass = _interopRequireDefault(require("./addToClass"));

var _addToFunctionOrVar = _interopRequireDefault(require("./addToFunctionOrVar"));

var _extractTypeProperties = _interopRequireDefault(require("./extractTypeProperties"));

var _upsertImport = _interopRequireDefault(require("./upsertImport"));

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

/**
 * @copyright   2018-2019, Miles Johnson
 * @license     https://opensource.org/licenses/MIT
 */
const BABEL_VERSION = 7;
const MAX_DEPTH = 3;
const MAX_SIZE = 25;
const REACT_FC_NAMES = ['SFC', 'StatelessComponent', 'FC', 'FunctionComponent'];

function isNotTS(name) {
  return name.endsWith('.js') || name.endsWith('.jsx');
}

var _default = (0, _helperPluginUtils.declare)((api, options, root) => {
  api.assertVersion(BABEL_VERSION);
  return {
    inherits: _pluginSyntaxTypescript.default,

    manipulateOptions(opts, parserOptions) {
      parserOptions.plugins.push('jsx');
    },

    post() {
      this.state = null;
    },

    pre() {
      this.state = {
        airbnbPropTypes: {
          count: 0,
          forbidImport: '',
          hasImport: false,
          namedImports: []
        },
        componentTypes: {},
        filePath: '',
        options: {
          customPropTypeSuffixes: [],
          forbidExtraProps: false,
          maxDepth: MAX_DEPTH,
          maxSize: MAX_SIZE,
          typeCheck: false,
          ...options
        },
        propTypes: {
          count: 0,
          defaultImport: '',
          hasImport: false
        },
        reactImportedName: '',
        referenceTypes: {}
      };
    },

    visitor: {
      Program: {
        enter(programPath, {
          filename
        }) {
          const state = this.state;
          state.filePath = filename;

          if (isNotTS(filename)) {
            return;
          }

          programPath.node.body.forEach(node => {
            if (!_core.types.isImportDeclaration(node)) {
              return;
            }

            if (node.source.value === 'prop-types') {
              const response = (0, _upsertImport.default)(node, {
                checkForDefault: 'PropTypes'
              });
              state.propTypes.hasImport = true;
              state.propTypes.defaultImport = response.defaultImport;
            }

            if (node.source.value === 'airbnb-prop-types') {
              const response = (0, _upsertImport.default)(node, {
                checkForNamed: 'forbidExtraProps'
              });
              state.airbnbPropTypes.hasImport = true;
              state.airbnbPropTypes.namedImports = response.namedImports;
              state.airbnbPropTypes.forbidImport = response.namedImport;
            }

            if (node.source.value === 'react') {
              const response = (0, _upsertImport.default)(node, {
                checkForDefault: 'React'
              });
              state.reactImportedName = response.defaultImport;
            }
          });

          if (!state.propTypes.hasImport && state.reactImportedName) {
            state.propTypes.defaultImport = (0, _helperModuleImports.addDefault)(programPath, 'prop-types', {
              nameHint: 'pt'
            }).name;
          }

          if (!state.airbnbPropTypes.hasImport && state.reactImportedName && options.forbidExtraProps) {
            state.airbnbPropTypes.forbidImport = (0, _helperModuleImports.addNamed)(programPath, 'forbidExtraProps', 'airbnb-prop-types').name;
            state.airbnbPropTypes.count += 1;
          }

          if (!state.propTypes.defaultImport && !state.reactImportedName) {
            return;
          }

          programPath.traverse({
            CallExpression({
              node
            }) {
              const {
                namedImports
              } = state.airbnbPropTypes;

              if (options.forbidExtraProps && _core.types.isIdentifier(node.callee) && namedImports.includes(node.callee.name)) {
                state.airbnbPropTypes.count += 1;
              }
            },

            'ClassDeclaration|ClassExpression': path => {
              const {
                node
              } = path;

              const valid = node.superTypeParameters && (_core.types.isMemberExpression(node.superClass) && _core.types.isIdentifier(node.superClass.object, {
                name: state.reactImportedName
              }) && (_core.types.isIdentifier(node.superClass.property, {
                name: 'Component'
              }) || _core.types.isIdentifier(node.superClass.property, {
                name: 'PureComponent'
              })) || state.reactImportedName && (_core.types.isIdentifier(node.superClass, {
                name: 'Component'
              }) || _core.types.isIdentifier(node.superClass, {
                name: 'PureComponent'
              })));

              if (valid) {
                (0, _addToClass.default)(node, state);
              }
            },

            FunctionDeclaration(path) {
              const {
                node
              } = path;
              const param = node.params[0];
              const valid = param && state.reactImportedName && node.id.name.match(/^[A-Z]/u) && (_core.types.isIdentifier(param) && param.typeAnnotation || _core.types.isObjectPattern(param) && param.typeAnnotation);

              if (valid) {
                (0, _addToFunctionOrVar.default)(path, node.id.name, state);
              }
            },

            Identifier({
              node
            }) {
              const {
                namedImports
              } = state.airbnbPropTypes;

              if (options.forbidExtraProps && namedImports.includes(node.name)) {
                state.airbnbPropTypes.count += 1;
              }
            },

            MemberExpression({
              node
            }) {
              if (_core.types.isIdentifier(node.object, {
                name: state.propTypes.defaultImport
              })) {
                state.propTypes.count += 1;
              }
            },

            TSInterfaceDeclaration({
              node
            }) {
              state.componentTypes[node.id.name] = (0, _extractTypeProperties.default)(node, state.componentTypes);
              state.referenceTypes[node.id.name] = node;
            },

            TSTypeAliasDeclaration({
              node
            }) {
              state.componentTypes[node.id.name] = (0, _extractTypeProperties.default)(node, state.componentTypes);
              state.referenceTypes[node.id.name] = node;
            },

            VariableDeclaration(path) {
              const {
                node
              } = path;

              if (node.declarations.length === 0) {
                return;
              }

              const id = node.declarations[0].id;

              if (!id.typeAnnotation || !id.typeAnnotation.typeAnnotation) {
                return;
              }

              const type = id.typeAnnotation.typeAnnotation;
              const valid = _core.types.isTSTypeReference(type) && type.typeParameters && type.typeParameters.params.length > 0 && (_core.types.isTSQualifiedName(type.typeName) && _core.types.isIdentifier(type.typeName.left, {
                name: state.reactImportedName
              }) && REACT_FC_NAMES.some(name => _core.types.isIdentifier(type.typeName.right, {
                name
              })) || state.reactImportedName && REACT_FC_NAMES.some(name => _core.types.isIdentifier(type.typeName, {
                name
              })));

              if (valid) {
                (0, _addToFunctionOrVar.default)(path, id.name, state);
              }
            }

          });
        },

        exit(path, {
          filename
        }) {
          const state = this.state;

          if (isNotTS(filename)) {
            return;
          }

          path.get('body').forEach(bodyPath => {
            if (state.propTypes.count === 0 && _core.types.isImportDeclaration(bodyPath.node) && bodyPath.node.source.value === 'prop-types') {
              bodyPath.remove();
            }
          });
        }

      }
    }
  };
});

exports.default = _default;